from typing import Generator, Iterable
from pyspark.accumulators import Accumulator

LEN_ERROR = ValueError('x len should be equal to 2')


class Ccf:
    class Iterate:
        @staticmethod
        def _map(key: int, value: int) -> tuple:
            return ((key, value), (value, key))

        @staticmethod
        def _reduce(key: int,
                    values: Iterable[int],
                    accumulator: Accumulator = None) -> Generator:
            value_list = []
            min_value = key
            for value in values:
                if value < min_value:
                    min_value = value
                value_list.append(value)
            if (min_value < key):
                yield (key, min_value)
                for value in value_list:
                    if min_value != value:
                        yield (value, min_value)
                        if accumulator:
                            accumulator.add(1)

        @staticmethod
        def map(x: Iterable) -> tuple:
            if len(x) != 2:
                raise LEN_ERROR
            return __class__._map(*x)

        @staticmethod
        def reduce(x: Iterable, accumulator: Accumulator = None) -> Generator:
            if len(x) != 2:
                raise LEN_ERROR
            yield from __class__._reduce(*x, accumulator)

    class IterateSecondarySorting(Iterate):
        @staticmethod
        def _reduce(key: int,
                    values: Iterable[int],
                    accumulator: Accumulator = None) -> Generator:
            min_value = values[0]
            if min_value < key:
                yield (key, min_value)
                for value in values[1:]:
                    yield (value, min_value)
                    if accumulator:
                        accumulator.add(1)

        @staticmethod
        def reduce(x: Iterable, accumulator: Accumulator = None) -> Generator:
            if len(x) != 2:
                raise LEN_ERROR
            yield from __class__._reduce(*x, accumulator)

    class Dedup:
        @staticmethod
        def _map(key: int, value: int) -> tuple:
            return ((key, value), None)

        @staticmethod
        def _reduce(key: Iterable, value: object) -> tuple:
            return (key[0], key[1])

        @staticmethod
        def map(x: Iterable) -> tuple:
            if len(x) != 2:
                raise LEN_ERROR
            return __class__._map(*x)

        @staticmethod
        def reduce(x: Iterable) -> tuple:
            if len(x) != 2:
                raise LEN_ERROR
            return __class__._reduce(*x)



import networkx as nx


class GraphGenerator:
    def __init__(self, nodes: int, edges: int):
        self.graph = self._generate_random_graph(nodes, edges)

    def generate_new_random_graph(self, nodes: int, edges: int) -> None:
        self.graph = self._generate_random_graph(nodes, edges)

    def _generate_random_graph(self, nodes: int, edges: int) -> nx.Graph:
        return nx.gnm_random_graph(nodes, edges)

    def remove_null_degres_nodes(self) -> None:
        copy = self.graph.copy()
        self.graph = nx.subgraph_view(
            copy,
            filter_node=lambda node: bool(nx.degree(copy, node)),
        )

    def save(self, path: str) -> None:
        with open(path, 'wb') as f:
            nx.write_edgelist(self.graph, f)

    def draw(self):
        nx.draw(self.graph, with_labels=True)

    @staticmethod
    def generate_ccf_random_graph(nodes: int, edges: int):
        graph = GraphGenerator(nodes, edges)
        graph.remove_null_degres_nodes()
        return graph

    @property
    def number_connected_components(self) -> int:
        return nx.number_connected_components(self.graph)

    @property
    def edges(self) -> list:
        return self.graph.edges

    @property
    def nodes(self) -> list:
        return self.graph.nodes



from networkx.classes.graph import Graph
from pyspark import SparkContext
from ccf_spark.graph_generator import GraphGenerator
from ccf_spark.ccf import Ccf

CCF_DEDUP = Ccf.Dedup
CCF_ITERATE = Ccf.Iterate
CCF_ITERATE_SECONDARY_SORTING = Ccf.IterateSecondarySorting


class CcfSpark:
    def __init__(self,
                 sc: SparkContext,
                 secondary_sorting: bool = False,
                 graph: Graph = None,
                 file_path: str = None,
                 separator: str = ' '):
        self.sc = sc
        self.iterator = CCF_ITERATE_SECONDARY_SORTING if secondary_sorting else CCF_ITERATE
        self.secondary_sorting = secondary_sorting
        if graph:
            self.graph = sc.parallelize(graph.edges)
        elif file_path:
            # File line format expected : <int>%separator%<int>
            self.graph = sc.textFile(file_path).map(
                lambda x: tuple(map(int,
                                    x.split(separator)[:2])))
        else:
            self.graph = sc.parallelize(
                GraphGenerator.generate_ccf_random_graph(500, 350).edges)
        self.iterated = False
        self._accumulator = self.sc.accumulator(0)

    def iterate(self, with_distinct: bool = False) -> int:
        self._accumulator.value = 0
        iterator = self.iterator  # To avoid SPARK-5063 error.
        self.graph = self.graph.flatMap(iterator.map).groupByKey()
        if self.secondary_sorting:
            self.graph = self.graph.map(lambda x: (x[0], sorted(x[1])))
        self.graph = self.graph.flatMap(
            lambda x, accumulator=self._accumulator: iterator.reduce(
                x, accumulator)).sortByKey()
        if with_distinct:
            self.graph = self.graph.distinct()
        else:
            self.graph = self.graph.map(CCF_DEDUP.map).groupByKey()
            self.graph = self.graph.map(CCF_DEDUP.reduce)
        return self._accumulator.value

    def iterate_all(self, with_distinct: bool = False) -> None:
        while True:
            new_pairs = self.iterate(with_distinct)
            if not new_pairs:
                break
        self.iterated = True

    def print(self) -> list:
        return self.graph.collect()

    def number_of_connected_components(self) -> int:
        """
            Can only be used after iterate_all()
        """
        if not self.iterated:
            self.iterate_all()
        return self.graph.map(lambda x: (x[1], 1)).reduceByKey(
            lambda a, b: (a + b)).count()



from setuptools import setup, find_packages
from os import path

dir = path.abspath(path.dirname(__file__))
with open(path.join(dir, 'README.md'), encoding='utf-8') as f:
    README = f.read()

setup(
    name='ccf_spark',
    version='0.3.1',
    description='A little CCF implementation in a Spark context with networkx',
    license='MIT',
    long_description_content_type='text/markdown',
    long_description=README,
    author='Théo Chennebault',
    author_email='theo.chennebault@le-cab-politique.fr',
    url='https://github.com/ErnestBidouille/ccf-spark',
    packages=find_packages(exclude=['tests*']),
    install_requires=['networkx==2.5', 'pyspark==3.0.1'],
)
